"""
This is a collection of deprecated modules.

- *aperture_20170523.py* \
    old aperture tracing methods

- *aprec.py*\
    interface to IRAF aperture records

- *ccdproc_mod.py*\
    modified ccdproc.CCDData class, deprecated because new CCD class implemented

- *pyreduce.py*\
    python version of REDUCE, deprecated

- *stella.py*\
    wrapper of **stella** (Ritter et al. 2014)
    I can't install it now

- normalization.py
    not needed now because new make_normflat() method
"""